==> senf-0.2.0-py2.py3-none-any.whl: senf/_fsnative.py <==
# -*- coding: utf-8 -*-
# Copyright 2016 Christoph Reiter
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

import os
import sys
import ctypes

from . import _winapi as winapi
from ._compat import text_type, PY3, PY2, url2pathname, urlparse, quote


is_win = os.name == "nt"
is_unix = not is_win
is_darwin = sys.platform == "darwin"


def _fsnative(text):
    if not isinstance(text, text_type):
        raise TypeError("%r needs to be a text type (%r)" % (text, text_type))

    if is_unix:
        # First we go to bytes so we can be sure we have a valid source.
        # Theoretically we should fail here in case we have a non-unicode
        # encoding. But this would make everything complicated and there is
        # no good way to handle a failure from the user side. Instead
        # fall back to utf-8 which is the most likely the right choice in
        # a mis-configured environment
        encoding = _encoding
        try:
            path = text.encode(encoding)
        except UnicodeEncodeError:
            path = text.encode("utf-8")
        if PY3:
            return path.decode(_encoding, "surrogateescape")
        return path
    else:
        return text


def _create_fsnative(type_):
    # a bit of magic to make fsnative(u"foo") and isinstance(path, fsnative)
    # work

    class meta(type):

        def __instancecheck__(self, instance):
            return isinstance(instance, type_)

        def __subclasscheck__(self, subclass):
            return issubclass(subclass, type_)

    class impl(object):
        """fsnative(text)

        Args:
            text (text): The text to convert to a path
        Returns:
            fsnative:
                The new path. Depending on the Python version and platform
                this is either `text` or `bytes`.
        Raises:
            TypeError: In case something other then `text` has been passed

        This type is a virtual base class for the real path type.
        Instantiating it returns an instance of the real path type
        and it overrides instance and subclass checks so that

        ::

            isinstance(fsnative(u"foo"), fsnative) == True
            issubclass(type(fsnative(u"foo")), fsnative) == True

        works as well.

        Can't fail.
        """

        def __new__(cls, text=u""):
            return _fsnative(text)

    new_type = meta("fsnative", (object,), dict(impl.__dict__))
    new_type.__module__ = "senf"
    return new_type


fsnative_type = text_type if is_win or PY3 else bytes
fsnative = _create_fsnative(fsnative_type)


def _get_encoding():
    """The encoding used for paths, argv, environ, stdout and stdin"""

    encoding = sys.getfilesystemencoding()
    if encoding is None:
        if is_darwin:
            return "utf-8"
        elif is_win:
            return "mbcs"
        else:
            return "ascii"
    return encoding

_encoding = _get_encoding()


def path2fsn(path):
    """
    Args:
        path (pathlike): The path to convert
    Returns:
        fsnative
    Raises:
        TypeError: In case the type can't be converted to a `fsnative`
        ValueError: In case conversion fails

    Returns a fsnative path for a path-like.

    If the passed in path is a `fsnative` path it simply returns it.
    This will not fail for a valid path (Either retrieved from stdlib functions
    or `argv` or `environ` or through the `fsnative` helper)
    """

    # allow mbcs str on py2+win and bytes on py3
    if PY2:
        if is_win:
            if isinstance(path, str):
                path = path.decode(_encoding)
        else:
            if isinstance(path, unicode):
                path = path.encode(_encoding)
    else:
        # TODO: If it ever gets added to Python we should call os.fspath() here
        if isinstance(path, bytes):
            path = path.decode(_encoding, "surrogateescape")
        elif is_unix and isinstance(path, str):
            # make sure we can encode it and this is not just some random
            # unicode string
            path.encode(_encoding, "surrogateescape")

    if not isinstance(path, fsnative_type):
        raise TypeError("path needs to be %s", fsnative_type.__name__)

    return path


def fsn2text(path):
    """
    Args:
        path (fsnative): The path to convert
    Returns:
        `text`
    Raises:
        TypeError: In case no `fsnative` has been passed

    Converts a path to text. This process is not reversible and should
    only be used for display purposes.

    On Python 3 the resulting `str` will not contain surrogates.

    This can't fail if a valid `fsnative` gets passed.
    """

    if not isinstance(path, fsnative_type):
        raise TypeError("path needs to be %s", fsnative_type.__name__)

    if fsnative_type is bytes:
        return path.decode(_encoding, "replace")
    else:
        if PY2 or is_win:
            return path
        else:
            return path.encode(
                _encoding, "surrogateescape").decode(_encoding, "replace")


def text2fsn(text):
    """
    Args:
        text (text): The text to convert
    Returns:
        `fsnative`

    Takes `text` and converts it to a `fsnative`. This operation is not
    reversible and can't fail.

    This is the same as calling ``fsnative(text)`` and available for
    consistency.
    """

    return fsnative(text)


def fsn2bytes(path, encoding):
    """
    Args:
        path (fsnative): The path to convert
        encoding (`str` or `None`): `None` if you don't care about Windows
    Returns:
        `bytes`
    Raises:
        TypeError: If no `fsnative` path is passed
        ValueError: On Windows if no valid encoding is passed or encoding fails

    Turns a path to bytes. If the path is not associated with an encoding
    the passed encoding is used (under Windows for example).
    """

    if not isinstance(path, fsnative_type):
        raise TypeError("path needs to be %s", fsnative_type.__name__)

    if is_win:
        if encoding is None:
            raise ValueError("invalid encoding %r" % encoding)
        try:
            return path.encode(encoding)
        except LookupError:
            raise ValueError("invalid encoding %r" % encoding)
    elif PY2:
        return path
    else:
        return path.encode(_encoding, "surrogateescape")


def bytes2fsn(data, encoding):
    """
    Args:
        data (bytes): The data to convert
        encoding (`str` or `None`): `None` if you don't care about Windows
    Returns:
        `fsnative`
    Raises:
        TypeError: If no `bytes` path is passed
        ValueError: On Windows if no valid encoding is passed or decoding fails

    Turns bytes to a path. If the path is not associated with an encoding
    the passed encoding is used (under Windows for example)
    """

    if not isinstance(data, bytes):
        raise TypeError("data needs to be bytes")

    if is_win:
        if encoding is None:
            raise ValueError("invalid encoding %r" % encoding)
        try:
            return data.decode(encoding)
        except LookupError:
            raise ValueError("invalid encoding %r" % encoding)
    elif PY2:
        return data
    else:
        return data.decode(_encoding, "surrogateescape")


def uri2fsn(uri):
    """
    Args:
        uri (`text` or :obj:`python:str`): A file URI
    Returns:
        `fsnative`
    Raises:
        TypeError: In case an invalid type is passed
        ValueError: In case the URI isn't a valid file URI

    Takes a file URI and returns a fsnative path
    """

    if PY2:
        if isinstance(uri, unicode):
            uri = uri.encode("utf-8")
        if not isinstance(uri, bytes):
            raise TypeError("uri needs to be ascii str or unicode")
    else:
        if not isinstance(uri, str):
            raise TypeError("uri needs to be str")

    parsed = urlparse(uri)
    scheme = parsed.scheme
    netloc = parsed.netloc
    path = parsed.path

    if scheme != "file":
        raise ValueError("Not a file URI")

    if is_win:
        path = url2pathname(netloc + path)
        if netloc:
            path = "\\\\" + path
        if PY2:
            path = path.decode("utf-8")
        return path
    else:
        if PY2:
            return url2pathname(path)
        else:
            return fsnative(url2pathname(path))


def fsn2uri(path):
    """
    Args:
        path (fsnative): The path to convert to an URI
    Returns:
        `fsnative`
    Raises:
        TypeError: If no `fsnative` was passed
        ValueError: If the path can't be converted

    Takes a fsnative path and returns a file URI.

    On Windows this returns a unicode URI. If you want an ASCII URI
    use :func:`fsn2uri_ascii` instead.
    """

    if not isinstance(path, fsnative_type):
        raise TypeError("path needs to be %s", fsnative_type.__name__)

    if is_win:
        buf = ctypes.create_unicode_buffer(winapi.INTERNET_MAX_URL_LENGTH)
        length = winapi.DWORD(winapi.INTERNET_MAX_URL_LENGTH)
        flags = 0
        try:
            winapi.UrlCreateFromPathW(path, buf, ctypes.byref(length), flags)
        except WindowsError as e:
            raise ValueError(e)
        return buf[:length.value]
    else:
        if PY2:
            return "file://" + quote(path)
        else:
            return "file://" + quote(path.encode(_encoding, "surrogateescape"))


def fsn2uri_ascii(path):
    """
    Args:
        path (fsnative): The path to convert to an URI
    Returns:
        `str`: An ASCII only `str`
    Raises:
        TypeError: If no `fsnative` was passed
        ValueError: If the path can't be converted

    Takes a fsnative path and returns a file URI.

    Like fsn2uri() but returns ASCII only. On Windows non-ASCII characters
    will be encoded using utf-8 and then percent encoded.
    """

    if is_win:
        uri = fsn2uri(path)

        def quoter(c):
            try:
                c.encode("ascii")
            except ValueError:
                return quote(c.encode("utf-8"))
            else:
                return c

        uri = map(quoter, uri)
        if PY2:
            return u"".join(uri).encode("ascii")
        else:
            return "".join(uri)
    else:
        return fsn2uri(path)

==> senf-0.3.0-py2.py3-none-any.whl: senf/_fsnative.py <==
# -*- coding: utf-8 -*-
# Copyright 2016 Christoph Reiter
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

import os
import sys
import ctypes

from . import _winapi as winapi
from ._compat import text_type, PY3, PY2, url2pathname, urlparse, quote, \
    unquote


is_win = os.name == "nt"
is_unix = not is_win
is_darwin = sys.platform == "darwin"


def _fsnative(text):
    if not isinstance(text, text_type):
        raise TypeError("%r needs to be a text type (%r)" % (text, text_type))

    if is_unix:
        # First we go to bytes so we can be sure we have a valid source.
        # Theoretically we should fail here in case we have a non-unicode
        # encoding. But this would make everything complicated and there is
        # no good way to handle a failure from the user side. Instead
        # fall back to utf-8 which is the most likely the right choice in
        # a mis-configured environment
        encoding = _encoding
        try:
            path = text.encode(encoding)
        except UnicodeEncodeError:
            path = text.encode("utf-8")
        if PY3:
            return path.decode(_encoding, "surrogateescape")
        return path
    else:
        return text


def _create_fsnative(type_):
    # a bit of magic to make fsnative(u"foo") and isinstance(path, fsnative)
    # work

    class meta(type):

        def __instancecheck__(self, instance):
            # XXX: invalid str on Unix + Py3 still returns True here, but
            # might fail when passed to fsnative API. We could be more strict
            # here and call _validate_fsnative(), but then we could
            # have a value not being an instance of fsnative, while its type
            # is still a subclass of fsnative.. and this is enough magic
            # already.
            return isinstance(instance, type_)

        def __subclasscheck__(self, subclass):
            return issubclass(subclass, type_)

    class impl(object):
        """fsnative(text=u"")

        Args:
            text (text): The text to convert to a path
        Returns:
            fsnative: The new path.
        Raises:
            TypeError: In case something other then `text` has been passed

        This type is a virtual base class for the real path type.
        Instantiating it returns an instance of the real path type and it
        overrides instance and subclass checks so that `isinstance` and
        `issubclass` checks work:

        ::

            isinstance(fsnative(u"foo"), fsnative) == True
            issubclass(type(fsnative(u"foo")), fsnative) == True

        The real returned type is:

        - :obj:`python:unicode` under Python 2 + Windows
        - :obj:`python:str` under Python 2 + Unix
        - :obj:`python3:str` under Python 3 + Windows
        - :obj:`python3:str` + ``surrogates`` (only containing code points
          which can be encoded with the locale encoding) under Python 3 +
          Unix

        Constructing a `fsnative` can't fail.
        """

        def __new__(cls, text=u""):
            return _fsnative(text)

    new_type = meta("fsnative", (object,), dict(impl.__dict__))
    new_type.__module__ = "senf"
    return new_type


fsnative_type = text_type if is_win or PY3 else bytes
fsnative = _create_fsnative(fsnative_type)


def _validate_fsnative(path):
    """
    Args:
        path (fsnative)
    Returns:
        `text` on Windows, `bytes` on Unix
    Raises:
        TypeError: in case the type is wrong or the ´str` on Py3 + Unix
            can't be converted to `bytes`

    This helper allows to validate the type and content of a path.
    To reduce overhead the encoded value for Py3 + Unix is returned so
    it can be reused.
    """

    if not isinstance(path, fsnative_type):
        raise TypeError("path needs to be %s, not %s" % (
            fsnative_type.__name__, type(path).__name__))

    if PY3 and is_unix:
        try:
            return path.encode(_encoding, "surrogateescape")
        except UnicodeEncodeError:
            # This look more like ValueError, but raising only one error
            # makes things simpler... also one could say str + surrogates
            # is its own type
            raise TypeError("path contained Unicode code points not valid in"
                            "the current path encoding. To create a valid "
                            "path from Unicode use text2fsn()")

    return path


def _get_encoding():
    """The encoding used for paths, argv, environ, stdout and stdin"""

    encoding = sys.getfilesystemencoding()
    if encoding is None:
        if is_darwin:
            return "utf-8"
        elif is_win:
            return "mbcs"
        else:
            return "ascii"
    return encoding

_encoding = _get_encoding()


def path2fsn(path):
    """
    Args:
        path (pathlike): The path to convert
    Returns:
        `fsnative`
    Raises:
        TypeError: In case the type can't be converted to a `fsnative`
        ValueError: In case conversion fails

    Returns a `fsnative` path for a `pathlike`.
    """

    # allow mbcs str on py2+win and bytes on py3
    if PY2:
        if is_win:
            if isinstance(path, str):
                path = path.decode(_encoding)
        else:
            if isinstance(path, unicode):
                path = path.encode(_encoding)
    else:
        path = getattr(os, "fspath", lambda x: x)(path)
        if isinstance(path, bytes):
            path = path.decode(_encoding, "surrogateescape")
        elif is_unix and isinstance(path, str):
            # make sure we can encode it and this is not just some random
            # unicode string
            path.encode(_encoding, "surrogateescape")

    if not isinstance(path, fsnative_type):
        raise TypeError("path needs to be %s", fsnative_type.__name__)

    return path


def fsn2text(path):
    """
    Args:
        path (fsnative): The path to convert
    Returns:
        `text`
    Raises:
        TypeError: In case no `fsnative` has been passed

    Converts a `fsnative` path to `text`.

    This process is not reversible and should only be used for display
    purposes.
    """

    path = _validate_fsnative(path)

    if is_win:
        return path
    else:
        return path.decode(_encoding, "replace")


def text2fsn(text):
    """
    Args:
        text (text): The text to convert
    Returns:
        `fsnative`
    Raises:
        TypeError: In case no `text` has been passed

    Takes `text` and converts it to a `fsnative`.

    This operation is not reversible and can't fail.
    """

    return fsnative(text)


def fsn2bytes(path, encoding):
    """
    Args:
        path (fsnative): The path to convert
        encoding (`str` or `None`): `None` if you don't care about Windows
    Returns:
        `bytes`
    Raises:
        TypeError: If no `fsnative` path is passed
        ValueError: If encoding fails or no encoding is given

    Turns a `fsnative` path to `bytes`.

    The passed *encoding* is only used on platforms where paths are not
    associated with an encoding (Windows for example). If you don't care about
    Windows you can pass `None`.
    """

    path = _validate_fsnative(path)

    if is_win:
        if encoding is None:
            raise ValueError("invalid encoding %r" % encoding)

        try:
            return path.encode(encoding)
        except LookupError:
            raise ValueError("invalid encoding %r" % encoding)
    else:
        return path


def bytes2fsn(data, encoding):
    """
    Args:
        data (bytes): The data to convert
        encoding (`str` or `None`): `None` if you don't care about Windows
    Returns:
        `fsnative`
    Raises:
        TypeError: If no `bytes` path is passed
        ValueError: If decoding fails or no encoding is given

    Turns `bytes` to a `fsnative` path.

    The passed *encoding* is only used on platforms where paths are not
    associated with an encoding (Windows for example). If you don't care about
    Windows you can pass `None`.
    """

    if not isinstance(data, bytes):
        raise TypeError("data needs to be bytes")

    if is_win:
        if encoding is None:
            raise ValueError("invalid encoding %r" % encoding)
        try:
            return data.decode(encoding)
        except LookupError:
            raise ValueError("invalid encoding %r" % encoding)
    elif PY2:
        return data
    else:
        return data.decode(_encoding, "surrogateescape")


def uri2fsn(uri):
    """
    Args:
        uri (`text` or :obj:`python:str`): A file URI
    Returns:
        `fsnative`
    Raises:
        TypeError: In case an invalid type is passed
        ValueError: In case the URI isn't a valid file URI

    Takes a file URI and returns a `fsnative` path
    """

    if PY2:
        if isinstance(uri, unicode):
            uri = uri.encode("utf-8")
        if not isinstance(uri, bytes):
            raise TypeError("uri needs to be ascii str or unicode")
    else:
        if not isinstance(uri, str):
            raise TypeError("uri needs to be str")

    parsed = urlparse(uri)
    scheme = parsed.scheme
    netloc = parsed.netloc
    path = parsed.path

    if scheme != "file":
        raise ValueError("Not a file URI")

    if is_win:
        path = url2pathname(netloc + path)
        if netloc:
            path = "\\\\" + path
        if PY2:
            path = path.decode("utf-8")
        return path
    else:
        if PY2:
            return url2pathname(path)
        else:
            return fsnative(url2pathname(path))


def fsn2uri(path):
    """
    Args:
        path (fsnative): The path to convert to an URI
    Returns:
        `str`: An ASCII only URI
    Raises:
        TypeError: If no `fsnative` was passed
        ValueError: If the path can't be converted

    Takes a `fsnative` path and returns a file URI.

    On Windows non-ASCII characters will be encoded using utf-8 and then
    percent encoded.
    """

    path = _validate_fsnative(path)

    def _quote_path(path):
        # RFC 2396
        return quote(path, "/:@&=+$,")

    if is_win:
        buf = ctypes.create_unicode_buffer(winapi.INTERNET_MAX_URL_LENGTH)
        length = winapi.DWORD(winapi.INTERNET_MAX_URL_LENGTH)
        flags = 0
        try:
            winapi.UrlCreateFromPathW(path, buf, ctypes.byref(length), flags)
        except WindowsError as e:
            raise ValueError(e)
        uri = buf[:length.value]

        # For some reason UrlCreateFromPathW escapes some chars outside of
        # ASCII and some not. Unquote and re-quote with utf-8.
        if PY3:
            # latin-1 maps code points directly to bytes, which is what we want
            uri = unquote(uri, "latin-1")
        else:
            # Python 2 does what we want by default
            uri = unquote(uri)

        return _quote_path(uri.encode("utf-8"))

    else:
        return "file://" + _quote_path(path)
